/**********************************************************************
 * AVM.C Copyright (C) University of Lincolnshire and Humberside 1999 *
 *--------------------------------------------------------------------*
 * AVM - A Virtual Machine                                            *
 *--------------------------------------------------------------------*
 * Version: 1.00;  Author: Damian G Walker;  Date: 16-Feb-1999        *
 **********************************************************************/

#define _AVM_C_

/* included headers ***************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "avm.h"

/* level 4 routines ***************************************************/

/* addition */
static void vm_add(registers *vm_reg, unsigned char operand) {
  int newvalue; /* result of addition */
  vm_reg->d = newvalue = vm_reg->d + operand;
  if(vm_reg->d) vm_reg->f &= ~(1 << VM_EQZ); else vm_reg->f |= 1 << VM_EQZ;
  if(vm_reg->d == newvalue) vm_reg->f &= ~(1 << VM_CAR); else 
    vm_reg->f |= 1 << VM_CAR;
}

/* logical and */
static void vm_and(registers *vm_reg, unsigned char operand) {
  vm_reg->d &= operand;
  if(vm_reg->d) vm_reg->f &= ~(1 << VM_EQZ); else vm_reg->f |= 1 << VM_EQZ;
}

/* load D */
static void vm_ldd(registers *vm_reg, unsigned char operand) {
  vm_reg->d = operand;
  if(vm_reg->d) vm_reg->f &= ~(1 << VM_EQZ); else vm_reg->f |= 1 << VM_EQZ;
}

/* subtraction */
static void vm_sub(registers *vm_reg, unsigned char operand) {
  int newvalue; /* result of addition */
  vm_reg->d = newvalue = vm_reg->d - operand;
  if(vm_reg->d) vm_reg->f &= ~(1 << VM_EQZ); else vm_reg->f |= 1 << VM_EQZ;
  if(vm_reg->d == newvalue) vm_reg->f &= ~(1 << VM_CAR); else 
    vm_reg->f |= 1 << VM_CAR;
}

/* logical or */
static void vm_ord(registers *vm_reg, unsigned char operand) {
  vm_reg->d |= operand;
  if(vm_reg->d) vm_reg->f &= ~(1 << VM_EQZ); else vm_reg->f |= 1 << VM_EQZ;
}

/* compare */
static void vm_cmp(registers *vm_reg, unsigned char operand) {
  if(vm_reg->d - operand)
    vm_reg->f &= ~(1 << VM_EQZ); else vm_reg->f |= 1 << VM_EQZ;
  if((signed)vm_reg->d - operand >= 0)
    vm_reg->f &= ~(1 << VM_CAR); else vm_reg->f |= 1 << VM_CAR;
}

/* exclusive or */
static void vm_xor(registers *vm_reg, unsigned char operand) {
  vm_reg->d ^= operand;
  if(vm_reg->d) vm_reg->f &= ~(1 << VM_EQZ); else vm_reg->f |= 1 << VM_EQZ;
}

/* jump if equal */
static void vm_jeq(registers *vm_reg, unsigned short address) {
  if(vm_reg->f & 1 << VM_EQZ) vm_reg->p = address & 0xfff;
}

/* jump unconditional */
static void vm_jmp(registers *vm_reg, unsigned short address) {
  vm_reg->p = address & 0xfff;
}

/* jump if carry */
static void vm_jca(registers *vm_reg, unsigned short address) {
  if(vm_reg->f & 1 << VM_CAR) vm_reg->p = address & 0xfff;
}

/* jump if not equal */
static void vm_jne(registers *vm_reg, unsigned short address) {
  if(!(vm_reg->f & 1 << VM_EQZ)) vm_reg->p = address & 0xfff;
}

/* jump if not carry */
static void vm_jnc(registers *vm_reg, unsigned short address) {
  if(!(vm_reg->f & 1 << VM_CAR)) vm_reg->p = address & 0xfff;
}

/* call subroutine */
static void vm_csr(registers *vm_reg, unsigned char vm_memspace[],
		   unsigned short address) {
  vm_memspace[vm_reg->s--] = vm_reg->p & 0xff;
  if((vm_reg->s &= 0xfff) > VM_STACK ) vm_reg->f |= 1 << VM_HLT;
  vm_memspace[vm_reg->s--] = vm_reg->p >> 8;
  if((vm_reg->s &= 0xfff) > VM_STACK ) vm_reg->f |= 1 << VM_HLT;
  vm_reg->p = address & 0xfff;
}

/* push A */
static void vm_pha(registers *vm_reg, unsigned char vm_memspace[]) {
  vm_memspace[vm_reg->s--] = vm_reg->a & 0xff;
  if((vm_reg->s &= 0xfff) > VM_STACK ) vm_reg->f |= 1 << VM_HLT;
  vm_memspace[vm_reg->s--] = vm_reg->a >> 8;
  if((vm_reg->s &= 0xfff) > VM_STACK ) vm_reg->f |= 1 << VM_HLT;
}

/* push D */
static void vm_phd(registers *vm_reg, unsigned char vm_memspace[]) {
  vm_memspace[vm_reg->s--] = vm_reg->d;
  if((vm_reg->s &= 0xfff) > VM_STACK ) vm_reg->f |= 1 << VM_HLT;
}

/* store D */
static void vm_std(unsigned char vm_memspace[], registers *vm_reg,
		   unsigned short address) {
  vm_memspace[address] = vm_reg->d;
}

/* pop A */
static void vm_ppa(registers *vm_reg, unsigned char vm_memspace[]) {
  vm_reg->a = (vm_memspace[++vm_reg->s] << 8) & 0xf00;
  vm_reg->a += vm_memspace[++vm_reg->s];
  if(vm_reg->a) vm_reg->f &= (1 << VM_EQZ) & 0xf; else
    vm_reg->f |= 1 << VM_EQZ;
  if(vm_reg->s > VM_STACK) vm_reg->f |= 1 << VM_HLT;
}

/* pop D */
static void vm_ppd(registers *vm_reg, unsigned char vm_memspace[]) {
  vm_reg->d = vm_memspace[++vm_reg->s];
  if(vm_reg->d) vm_reg->f &= (1 << VM_EQZ) & 0xf; else
    vm_reg->f |= 1 << VM_EQZ;
  if(vm_reg->s > VM_STACK) vm_reg->f |= 1 << VM_HLT;
}

/* return from subroutine */
static void vm_ret(registers *vm_reg, unsigned char vm_memspace[]) {
  vm_reg->p = (vm_memspace[++vm_reg->s] << 8) & 0xf00;
  vm_reg->p += vm_memspace[++vm_reg->s];
  if(vm_reg->s > VM_STACK) vm_reg->f |= 1 << VM_HLT;
}

/* load A register */
static void vm_lda(registers *vm_reg, unsigned int data) {
  vm_reg->a = data;
  if(vm_reg->a)
    vm_reg->f &= ~(1 << VM_EQZ) & 0xf; else vm_reg->f |= 1 << VM_EQZ;
}

/* store A register */
static void vm_sta(unsigned char vm_memspace[], registers *vm_reg, 
		   unsigned short address) {
  vm_memspace[address] = vm_reg->a >> 8;
  vm_memspace[address + 1] = vm_reg->a & 0xff;
}

/* increment A register */
static void vm_inc(registers *vm_reg) {
  vm_reg->a = (vm_reg->a + 1) & 0xfff;
  if(!vm_reg->a)
    vm_reg->f |= 1 << VM_CAR | 1 << VM_EQZ;
  else
    vm_reg->f &= ~(1 << VM_CAR | 1 << VM_EQZ);
}

/* decrement A register */
static void vm_dec(registers *vm_reg) {
  if(!vm_reg->a) vm_reg->f |= 1 << VM_CAR; else vm_reg->f &= ~(1 << VM_CAR);
  vm_reg->a = (vm_reg->a - 1) & 0xfff;
  if(!vm_reg->a) vm_reg->f |= 1 << VM_EQZ; else vm_reg->f &= ~(1 << VM_EQZ);
}

/* level 3 routines ***************************************************/

/* execute immediate instructions */
static void vm_execimmediate(registers *vm_reg, unsigned char vm_memspace[]) {
  switch( (vm_reg->i >> 8) & 0xf ) {
  case 0x0: vm_reg->f |= 1 << VM_HLT; break;
  case 0x1: vm_jeq(vm_reg, vm_reg->p + (signed char)(vm_reg->i & 0xff)); break;
  case 0x2: vm_jmp(vm_reg, vm_reg->p + (signed char)(vm_reg->i & 0xff)); break;
  case 0x3: vm_jca(vm_reg, vm_reg->p + (signed char)(vm_reg->i & 0xff)); break;
  case 0x4: vm_reg->f |= 1 << VM_HLT; break;
  case 0x5: vm_jne(vm_reg, vm_reg->p + (signed char)(vm_reg->i & 0xff)); break;
  case 0x6: 
    vm_csr(vm_reg, vm_memspace, vm_reg->p + (signed char)(vm_reg->i & 0xff)); 
    break;
  case 0x7: vm_jnc(vm_reg, vm_reg->p + (signed char)(vm_reg->i & 0xff)); break;
  case 0x8: vm_add(vm_reg, vm_reg->i & 0xff); break;
  case 0x9: vm_and(vm_reg, vm_reg->i & 0xff); break;
  case 0xa: vm_reg->f |= 1 << VM_HLT; break;
  case 0xb: 
    vm_ldd(vm_reg, vm_reg->i & 0xff); 
    break;
  case 0xc: vm_sub(vm_reg, vm_reg->i & 0xff); break;
  case 0xd: vm_ord(vm_reg, vm_reg->i & 0xff); break;
  case 0xe: vm_cmp(vm_reg, vm_reg->i & 0xff); break;
  case 0xf: vm_xor(vm_reg, vm_reg->i & 0xff); break;
  }
}

/* execute implied instructions */
static void vm_execimplied(registers *vm_reg, unsigned char vm_memspace[]) {
  switch(vm_reg->i & 0xf) {
  case 0x0: vm_pha(vm_reg, vm_memspace); break;
  case 0x1: vm_phd(vm_reg, vm_memspace); break;
  case 0x2: vm_inc(vm_reg); break;
  case 0x3: vm_std(vm_memspace, vm_reg, vm_reg->a); break;
  case 0x4: vm_ppa(vm_reg, vm_memspace); break;
  case 0x5: vm_ppd(vm_reg, vm_memspace); break;
  case 0x6: vm_ret(vm_reg, vm_memspace); break;
  case 0x7: vm_reg->f |= 1 << VM_HLT; break;
  case 0x8: vm_add(vm_reg, vm_memspace[vm_reg->a]); break;
  case 0x9: vm_and(vm_reg, vm_memspace[vm_reg->a]); break;
  case 0xa: vm_dec(vm_reg); break;
  case 0xb: vm_ldd(vm_reg, vm_memspace[vm_reg->a]); break;
  case 0xc: vm_sub(vm_reg, vm_memspace[vm_reg->a]); break;
  case 0xd: vm_ord(vm_reg, vm_memspace[vm_reg->a]); break;
  case 0xe: vm_cmp(vm_reg, vm_memspace[vm_reg->a]); break;
  case 0xf: vm_xor(vm_reg, vm_memspace[vm_reg->a]); break;
  }
}

/* execute indirect instructions */
static void vm_execindirect(registers *vm_reg, unsigned char vm_memspace[]) {
  switch(vm_reg->i >> 12) {
  case 0x0: vm_reg->f |= 1 << VM_HLT; break;
  case 0x1: 
    vm_lda(vm_reg, ((vm_memspace[vm_reg->i & 0xfff] << 8) +
      vm_memspace[(vm_reg->i & 0xfff) + 1]) & 0xfff);
    break;
  case 0x2: vm_jmp(vm_reg, vm_reg->i & 0xfff); break;
  case 0x3: vm_std(vm_memspace, vm_reg, vm_reg->i & 0xfff); break;
  case 0x4: vm_lda(vm_reg, vm_reg->i & 0xfff); break;
  case 0x5: vm_sta(vm_memspace, vm_reg, vm_reg->i & 0xfff); break;
  case 0x6: vm_csr(vm_reg, vm_memspace, vm_reg->i & 0xfff); break;
  case 0x7: vm_reg->f |= 1 << VM_HLT; break;
  case 0x8: vm_add(vm_reg, vm_memspace[vm_reg->i & 0xfff]); break;
  case 0x9: vm_and(vm_reg, vm_memspace[vm_reg->i & 0xfff]); break;
  case 0xa: vm_reg->f |= 1 << VM_HLT; break;
  case 0xb: vm_ldd(vm_reg, vm_memspace[vm_reg->i & 0xfff]); break;
  case 0xc: vm_sub(vm_reg, vm_memspace[vm_reg->i & 0xfff]); break;
  case 0xd: vm_ord(vm_reg, vm_memspace[vm_reg->i & 0xfff]); break;
  case 0xe: vm_cmp(vm_reg, vm_memspace[vm_reg->i & 0xfff]); break;
  case 0xf: vm_xor(vm_reg, vm_memspace[vm_reg->i & 0xfff]); break;
  }
}

/* level 2 routines ***************************************************/

/* initialise the virtual machine */
static registers vm_initreg(void) {
  registers vm_reg; /* register set */
  vm_reg.a = vm_reg.d = vm_reg.f = 0;
  vm_reg.p = VM_START;
  vm_reg.s = VM_STACK;
  return vm_reg;
}

/* initialise the disk drive */
static diskdrive vm_initdisk(void) {
  diskdrive vm_disk; /* disk drive */
  int bcount; /* byte count */
  if(!( vm_disk.diskfile = fopen("avm.dsk", "r+b") ))
    if(( vm_disk.diskfile = fopen("avm.dsk", "w+b") ))
      for(bcount = 0; bcount < 65536; ++bcount)
	fputc(0, vm_disk.diskfile);
  vm_disk.pos = vm_disk.sector = 0;
  return vm_disk;
}

/* execute top-level instruction */
static void vm_exec(registers *vm_reg, unsigned char vm_memspace[]) {
  vm_reg->i = vm_memspace[vm_reg->p++];
  switch(vm_reg->i >> 4) {
  case 0x0: /* immediate and relative addressing */
    vm_reg->i = (vm_reg->i << 8) + vm_memspace[vm_reg->p++];
    vm_execimmediate(vm_reg, vm_memspace);
    break;
  case 0x7: /* implied addressing */
    vm_execimplied(vm_reg, vm_memspace);
    break;
  case 0x1: /* indirect addressing */
  case 0x2: 
  case 0x3: 
  case 0x4: 
  case 0x5: 
  case 0x6: 
  case 0x8: 
  case 0x9: 
  case 0xa: 
  case 0xb: 
  case 0xc: 
  case 0xd: 
  case 0xe: 
  case 0xf: 
    vm_reg->i = (vm_reg->i << 8) + vm_memspace[vm_reg->p++];
    vm_execindirect(vm_reg, vm_memspace);
    break;
  default: /* anything unrecognised */
    vm_reg->f |= (1 << VM_HLT);
  }
  vm_reg->p &= 0xfff;
  vm_reg->a &= 0xfff;
  vm_reg->s &= 0xfff;
}

/* perform input and output */
static void vm_io(unsigned char vm_memspace[], diskdrive *vm_disk) {
  if( vm_memspace[VM_IOCTL] & VM_IOCINP )
    vm_memspace[VM_IODATA] = getchar();
  if( vm_memspace[VM_IOCTL] & VM_IOCPRT )
    putchar( vm_memspace[VM_IODATA] );
  if( vm_memspace[VM_IOCTL] & VM_IOCSEC )
    vm_disk->sector = vm_memspace[VM_IODATA];
  if( vm_memspace[VM_IOCTL] & VM_IOCPOS )
    vm_disk->pos = vm_memspace[VM_IODATA];
  if( (vm_memspace[VM_IOCTL] & VM_IOCRED) && vm_disk->diskfile )
     vm_memspace[VM_IODATA] = fgetc(vm_disk->diskfile);
  if( (vm_memspace[VM_IOCTL] & VM_IOCWRT) && vm_disk->diskfile )
    fputc( vm_memspace[VM_IODATA], vm_disk->diskfile );
  if( vm_memspace[VM_IOCTL] & VM_IOCADV )
    if(!( vm_disk->pos = (vm_disk->pos + 1) & 0xff ))
      vm_disk->sector = (vm_disk->sector + 1) & 0xff;
  if(vm_disk->diskfile)
    fseek(vm_disk->diskfile, (vm_disk->sector << 8) + vm_disk->pos, SEEK_SET);
  vm_memspace[VM_IOCTL] = 0;
}

/* close the virtual machine */
static void vm_closedisk(diskdrive vm_disk) {
  if(vm_disk.diskfile) fclose(vm_disk.diskfile);
}

/* level 1 routines ***************************************************/

/* load an individual code file */
int loadfile(unsigned char vm_memspace[], char *filename) {
  FILE *progfile; /* handle for program file */
  unsigned int org, /* origin address */
    filelen; /* length of data file */
  int inp; /* input character */
  filelen = 0;
  if(!( progfile = fopen(filename, "rb") )) return 0;
  fprintf(stderr, "Loading %s: ", filename);
  org = fgetc(progfile) << 4;
  fprintf(stderr, "origin $%03x... ", org);
  while(( inp = fgetc(progfile) ) != EOF) {
    vm_memspace[org] = inp;
    org = (org + 1) & 0xfff;
    ++filelen;
  }
  fprintf(stderr, "%d bytes read\n", filelen);
  fclose(progfile);
  return !0;
}

/* start the virtual machine */
void vmstart(unsigned char vm_memspace[]) {
  registers vm_reg; /* vm's registers */
  diskdrive vm_disk; /* vm's disk drive */
  vm_reg = vm_initreg();
  vm_disk = vm_initdisk();
  while(!( vm_reg.f & (1 << VM_HLT) )) {
    if( vm_memspace[VM_DEBUG] & 1 )
      printf("A=$%03x D=$%02x F=$%1x I=$%04x P=$%03x S=$%03x\n",
	     vm_reg.a, vm_reg.d, vm_reg.f, vm_reg.i, vm_reg.p, vm_reg.s);
    vm_exec(&vm_reg, vm_memspace);
    vm_io(vm_memspace, &vm_disk);
  }
  if( vm_memspace[VM_DEBUG] & 1 ) {
    puts("---");
    printf("A=$%03x D=$%02x F=$%1x I=$%04x P=$%03x S=$%03x\n",
	   vm_reg.a, vm_reg.d, vm_reg.f, vm_reg.i, vm_reg.p, vm_reg.s);
  }
  vm_closedisk(vm_disk);
}

/* level 0 routines ***************************************************/

int main(int argc, char **argv) {
  static unsigned char vm_memspace[VM_MEMSIZE]; /* vm's memory */
  int argn; /* used to count through arguments */
  for(argn = 1; argn < argc; ++argn) {
    if( !strcmp(argv[ argn ], "-d") || !strcmp(argv[ argn ], "--debug") )
      vm_memspace[VM_DEBUG] |= 1;
    else if( !loadfile(vm_memspace, argv[ argn ]) )
      return 1;
  }
  vmstart(vm_memspace);
  return 0;
}
