/**********************************************************************
 * AVM.H Copyright (C) University of Lincolnshire and Humberside 1999 *
 *--------------------------------------------------------------------*
 * AVM - Main Header File                                             *
 *--------------------------------------------------------------------*
 * Version: 1.00;  Author: Damian G Walker;  Date: 16-Feb-1999        *
 **********************************************************************/

#ifndef _AVM_H_
#define _AVM_H_

/* defined constants **************************************************/

#define VM_MEMSIZE 0x1000 /* size of memory */
#define VM_IODATA 0xfff /* data address of virtual serial I/O */
#define VM_IOCTL 0xffe /* control address for I/O */
#define VM_DEBUG 0xffd /* control address for debugger */
#define VM_START 0 /* address where VM execution starts */
#define VM_STACK 0xff0

#define VM_IOCINP 0x01 /* I/O control input command */
#define VM_IOCPRT 0x02 /* I/O control print command */
#define VM_IOCSEC 0x04 /* I/O control seek sector command */
#define VM_IOCPOS 0x08 /* I/O control seek within command */
#define VM_IOCRED 0x10 /* I/O control read byte command */
#define VM_IOCWRT 0x20 /* I/O control write byte command */
#define VM_IOCADV 0x40 /* I/O control disk advance command */

#define VM_HLT 0 /* halt flag bit */
#define VM_EQZ 1 /* equal/zero flag bit */
#define VM_RST 2 /* reset flag bit */
#define VM_CAR 3 /* carry flag bit */

/* data structures ****************************************************/

typedef struct { /* AVM register set */
  unsigned short a; /* 12-bit address register */
  unsigned char d; /* 8-bit data register */
  unsigned char f; /* 4-bit flags register */
  unsigned short i; /* 16-bit opcode/operand */
  unsigned short p; /* 12-bit program counter */
  unsigned short s; /* 12-bit stack pointer */
} registers;

typedef struct { /* AVM virtual disk drive */
  FILE *diskfile; /* file for disk I/O */
  unsigned int sector, /* sector number */
    pos; /* current position in sector */
} diskdrive;

/* end of header ******************************************************/

#endif
